"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const path = require("path");
const glob = require("glob");
const config_1 = require("./../modules/config");
const log_1 = require("../modules/log");
const utils_1 = require("../modules/utils");
const id_1 = require("./process/id");
const env_1 = require("./process/env");
const pid_1 = require("./process/pid");
const watchFileAndAct_1 = require("./process/watchFileAndAct");
const typescript_1 = require("./process/typescript");
let api;
let config = {};
class Process {
    constructor() {
        // Only in files required by `index.js` do we need to delay the loading of the API object
        // This is due to cyclical require issues
        api = require("../index").api;
        this.initializers = {};
        this.loadInitializers = [];
        this.startInitializers = [];
        this.stopInitializers = [];
        this.startCount = 0;
        api.commands.initialize = async (...args) => {
            return this.initialize(...args);
        };
        api.commands.start = async (...args) => {
            return this.start(...args);
        };
        api.commands.stop = async () => {
            return this.stop();
        };
        api.commands.restart = async () => {
            return this.restart();
        };
        api.process = this;
    }
    async initialize(params = {}) {
        this._startingParams = params;
        const loadInitializerRankings = {};
        const startInitializerRankings = {};
        const stopInitializerRankings = {};
        let initializerFiles = [];
        // rebuild config with startingParams
        config = config_1.buildConfig(this._startingParams);
        // load initializers from core
        initializerFiles = initializerFiles.concat(glob.sync(path.join(__dirname, "..", "initializers", "**", "**/*(*.js|*.ts)")));
        // load initializers from project
        config.general.paths.initializer.forEach((startPath) => {
            initializerFiles = initializerFiles.concat(glob.sync(path.join(startPath, "**", "**/*(*.js|*.ts)")));
        });
        // load initializers from plugins
        for (const pluginName in config.plugins) {
            if (config.plugins[pluginName] !== false) {
                const pluginPath = config.plugins[pluginName].path;
                // old style at the root of the project
                initializerFiles = initializerFiles.concat(glob.sync(path.join(pluginPath, "initializers", "**", "**/*(*.js|*.ts)")));
                // dist files if running in JS mode
                if (!typescript_1.typescript) {
                    initializerFiles = initializerFiles.concat(glob.sync(path.join(pluginPath, "dist", "initializers", "**", "*.js")));
                }
                // src files if running in TS mode
                if (typescript_1.typescript) {
                    initializerFiles = initializerFiles.concat(glob.sync(path.join(pluginPath, "src", "initializers", "**", "*.ts")));
                }
            }
        }
        initializerFiles = utils_1.utils.arrayUniqueify(initializerFiles);
        initializerFiles = utils_1.utils.ensureNoTsHeaderFiles(initializerFiles);
        initializerFiles.forEach(f => {
            const file = path.normalize(f);
            delete require.cache[require.resolve(file)];
            let exportedClasses = require(file);
            // allow for old-js style single default exports
            if (typeof exportedClasses === "function") {
                exportedClasses = { default: exportedClasses };
            }
            if (Object.keys(exportedClasses).length === 0) {
                this.fatalError(new Error(`no exported intializers found in ${file}`), file);
            }
            for (const exportKey in exportedClasses) {
                let initializer;
                let InitializerClass = exportedClasses[exportKey];
                try {
                    initializer = new InitializerClass();
                    // check if initializer already exists (exclude utils and config)
                    if (this.initializers[initializer.name]) {
                        const warningMessage = `an existing intializer with the same name \`${initializer.name}\` will be overridden by the file ${file}`;
                        log_1.log(warningMessage, "warning");
                    }
                    else {
                        initializer.validate();
                        this.initializers[initializer.name] = initializer;
                    }
                }
                catch (error) {
                    this.fatalError(error, file);
                }
                const initializeFunction = async () => {
                    watchFileAndAct_1.watchFileAndAct(file, async () => {
                        log_1.log(`*** Rebooting due to initializer change (${file}) ***`, "info");
                        await this.restart();
                    });
                    if (typeof initializer.initialize === "function") {
                        log_1.log(`Loading initializer: ${initializer.name}`, "debug", file);
                        try {
                            await initializer.initialize(config);
                            try {
                                log_1.log(`Loaded initializer: ${initializer.name}`, "debug", file);
                            }
                            catch (e) { }
                        }
                        catch (error) {
                            const message = `Exception occured in initializer \`${initializer.name}\` during load`;
                            try {
                                log_1.log(message, "emerg", error.toString());
                            }
                            catch (_error) {
                                console.error(message);
                            }
                            throw error;
                        }
                    }
                };
                const startFunction = async () => {
                    if (typeof initializer.start === "function") {
                        log_1.log(`Starting initializer: ${initializer.name}`, "debug", file);
                        try {
                            await initializer.start(config);
                            log_1.log(`Started initializer: ${initializer.name}`, "debug", file);
                        }
                        catch (error) {
                            log_1.log(`Exception occured in initializer \`${initializer.name}\` during start`, "emerg", error.toString());
                            throw error;
                        }
                    }
                };
                const stopFunction = async () => {
                    if (typeof initializer.stop === "function") {
                        log_1.log(`Stopping initializer: ${initializer.name}`, "debug", file);
                        try {
                            await initializer.stop(config);
                            log_1.log(`Stopped initializer: ${initializer.name}`, "debug", file);
                        }
                        catch (error) {
                            log_1.log(`Exception occured in initializer \`${initializer.name}\` during stop`, "emerg", error.toString());
                            throw error;
                        }
                    }
                };
                if (loadInitializerRankings[initializer.loadPriority] === undefined) {
                    loadInitializerRankings[initializer.loadPriority] = [];
                }
                if (startInitializerRankings[initializer.startPriority] === undefined) {
                    startInitializerRankings[initializer.startPriority] = [];
                }
                if (stopInitializerRankings[initializer.stopPriority] === undefined) {
                    stopInitializerRankings[initializer.stopPriority] = [];
                }
                if (initializer.loadPriority > 0) {
                    loadInitializerRankings[initializer.loadPriority].push(initializeFunction);
                }
                if (initializer.startPriority > 0) {
                    startInitializerRankings[initializer.startPriority].push(startFunction);
                }
                if (initializer.stopPriority > 0) {
                    stopInitializerRankings[initializer.stopPriority].push(stopFunction);
                }
            }
        });
        // flatten all the ordered initializer methods
        this.loadInitializers = this.flattenOrderedInitialzer(loadInitializerRankings);
        this.startInitializers = this.flattenOrderedInitialzer(startInitializerRankings);
        this.stopInitializers = this.flattenOrderedInitialzer(stopInitializerRankings);
        try {
            await utils_1.utils.asyncWaterfall(this.loadInitializers);
        }
        catch (error) {
            return this.fatalError(error, "initialize");
        }
        this.initialized = true;
        return api;
    }
    async start(params = {}) {
        if (this.initialized !== true) {
            await this.initialize(params);
        }
        pid_1.writePidFile();
        this.running = true;
        api.running = true;
        log_1.log(`environment: ${env_1.env}`, "notice");
        log_1.log("*** Starting ActionHero ***", "info");
        this.startInitializers.push(() => {
            this.bootTime = new Date().getTime();
            if (this.startCount === 0) {
                log_1.log(`server ID: ${id_1.id}`, "notice");
                log_1.log("*** ActionHero Started ***", "notice");
                this.startCount++;
            }
            else {
                log_1.log("*** ActionHero Restarted ***", "notice");
            }
        });
        try {
            await utils_1.utils.asyncWaterfall(this.startInitializers);
        }
        catch (error) {
            return this.fatalError(error, "start");
        }
        return api;
    }
    async stop() {
        if (this.running) {
            this.shuttingDown = true;
            this.running = false;
            this.initialized = false;
            log_1.log("stopping process...", "notice");
            await utils_1.utils.sleep(100);
            this.stopInitializers.push(async () => {
                pid_1.clearPidFile();
                log_1.log("*** ActionHero Stopped ***", "notice");
                delete this.shuttingDown;
                // reset initializers to prevent duplicate check on restart
                this.initializers = {};
                api.running = false;
                await utils_1.utils.sleep(100);
            });
            try {
                watchFileAndAct_1.unWatchAllFiles();
                await utils_1.utils.asyncWaterfall(this.stopInitializers);
            }
            catch (error) {
                return this.fatalError(error, "stop");
            }
        }
        else if (this.shuttingDown === true) {
            // double sigterm; ignore it
        }
        else {
            const message = "Cannot shut down actionhero, not running";
            log_1.log(message, "error");
        }
    }
    async restart() {
        if (this.running === true) {
            await this.stop();
            await this.start(this._startingParams);
        }
        else {
            await this.start(this._startingParams);
        }
        return api;
    }
    // HELPERS
    async fatalError(errors, type) {
        if (errors && !(errors instanceof Array)) {
            errors = [errors];
        }
        if (errors) {
            log_1.log(`Error with initializer step: ${JSON.stringify(type)}`, "emerg");
            errors.forEach(error => {
                log_1.log(error.stack, "emerg");
            });
            await this.stop();
            await utils_1.utils.sleep(1000); // allow time for console.log to print
            process.exit(1);
        }
    }
    flattenOrderedInitialzer(collection) {
        const output = [];
        const keys = [];
        for (const key in collection) {
            keys.push(parseInt(key));
        }
        keys.sort(sortNumber);
        keys.forEach(key => {
            collection[key].forEach(d => {
                output.push(d);
            });
        });
        return output;
    }
}
exports.Process = Process;
function sortNumber(a, b) {
    return a - b;
}
