"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const index_1 = require("./../index");
var task;
(function (task) {
    /**
     * Enqueue a task to be performed in the background.
     * Will throw an error if redis cannot be reached.
     */
    async function enqueue(taskName, params, queue = index_1.api.tasks.tasks[taskName].queue) {
        //@ts-ignore
        return index_1.api.resque.queue.enqueue(queue, taskName, params);
    }
    task.enqueue = enqueue;
    /**
     * Enqueue a task to be performed in the background, at a certain time in the future.
     * Will throw an error if redis cannot be reached.
     *
     * Inputs:
     * * taskName: The name of the task.
     * * params: Params to pass to the task.
     * * queue: (Optional) Which queue/priority to run this instance of the task on.
     */
    async function enqueueAt(timestamp, taskName, params, queue = index_1.api.tasks.tasks[taskName].queue) {
        //@ts-ignore
        return index_1.api.resque.queue.enqueueAt(timestamp, queue, taskName, params);
    }
    task.enqueueAt = enqueueAt;
    /**
     * Enqueue a task to be performed in the background, at a certain number of ms from now.
     * Will throw an error if redis cannot be reached.
     *
     * Inputs:
     * * timestamp: At what time the task is able to be run.  Does not gaurentee that the task will be run at this time. (in ms)
     * * taskName: The name of the task.
     * * params: Params to pass to the task.
     * * queue: (Optional) Which queue/priority to run this instance of the task on.
     */
    async function enqueueIn(time, taskName, params, queue = index_1.api.tasks.tasks[taskName].queue) {
        //@ts-ignore
        return index_1.api.resque.queue.enqueueIn(time, queue, taskName, params);
    }
    task.enqueueIn = enqueueIn;
    /**
     * Delete a previously enqueued task, which hasn't been run yet, from a queue.
     * Will throw an error if redis cannot be reached.
     *
     * Inputs:
     * * q: Which queue/priority is the task stored on?
     * * taskName: The name of the job, likley to be the same name as a tak.
     * * args: The arguments of the job.  Note, arguments passed to a Task initially may be modified when enqueuing.  It is best to read job properties first via `api.tasks.queued` or similar method.
     * * count: Of the jobs that match q, taskName, and args, up to what position should we delete? (Default 0; this command is 0-indexed)
     */
    async function del(q, taskName, args, count) {
        //@ts-ignore
        return index_1.api.resque.queue.del(q, taskName, args, count);
    }
    task.del = del;
    /**
     * Delete all previously enqueued tasks, which haven't been run yet, from all possible delayed timestamps.
     * Will throw an error if redis cannot be reached.
     *
     * Inputs:
     * * q: Which queue/priority is to run on?
     * * taskName: The name of the job, likley to be the same name as a tak.
     * * args  The arguments of the job.  Note, arguments passed to a Task initially may be modified when enqueuing. It is best to read job properties first via `api.tasks.delayedAt` or similar method.
     */
    async function delDelayed(q, taskName, args) {
        // @ts-ignore
        return index_1.api.resque.queue.delDelayed(q, taskName, args);
    }
    task.delDelayed = delDelayed;
    /**
     * Return the timestamps a task is scheduled for.
     * Will throw an error if redis cannot be reached.
     *
     * Inputs:
     * * q: Which queue/priority is to run on?
     * * taskName: The name of the job, likley to be the same name as a tak.
     * * args: The arguments of the job.  Note, arguments passed to a Task initially may be modified when enqueuing.  It is best to read job properties first via `api.tasks.delayedAt` or similar method.
     */
    async function scheduledAt(q, taskName, args) {
        //@ts-ignore
        return index_1.api.resque.queue.scheduledAt(q, taskName, args);
    }
    task.scheduledAt = scheduledAt;
    /**
     * Return all resque stats for this namespace (how jobs failed, jobs succeded, etc)
     * Will throw an error if redis cannot be reached.
     */
    async function stats() {
        return index_1.api.resque.queue.stats();
    }
    task.stats = stats;
    /**
     * Retrieve the details of jobs enqueued on a certain queue between start and stop (0-indexed)
     * Will throw an error if redis cannot be reached.
     *
     * Inputs:
     * * q      The name of the queue.
     * * start  The index of the first job to return.
     * * stop   The index of the last job to return.
     */
    async function queued(q, start, stop) {
        return index_1.api.resque.queue.queued(q, start, stop);
    }
    task.queued = queued;
    /**
     * Delete a queue in redis, and all jobs stored on it.
     * Will throw an error if redis cannot be reached.
     */
    async function delQueue(q) {
        return index_1.api.resque.queue.delQueue(q);
    }
    task.delQueue = delQueue;
    /**
     * Return any locks, as created by resque plugins or task middleware, in this redis namespace.
     * Will contain locks with keys like `resque:lock:{job}` and `resque:workerslock:{workerId}`
     * Will throw an error if redis cannot be reached.
     */
    async function locks() {
        return index_1.api.resque.queue.locks();
    }
    task.locks = locks;
    /**
     * Delete a lock on a job or worker.  Locks can be found via `api.tasks.locks`
     * Will throw an error if redis cannot be reached.
     */
    async function delLock(lock) {
        return index_1.api.resque.queue.delLock(lock);
    }
    task.delLock = delLock;
    /**
     * List all timestamps for which tasks are enqueued in the future, via `api.tasks.enqueueIn` or `api.tasks.enqueueAt`
     * Will throw an error if redis cannot be reached.
     */
    async function timestamps() {
        return index_1.api.resque.queue.timestamps();
    }
    task.timestamps = timestamps;
    /**
     * Return all jobs which have been enqueued to run at a certain timestamp.
     * Will throw an error if redis cannot be reached.
     */
    async function delayedAt(timestamp) {
        return index_1.api.resque.queue.delayedAt(timestamp);
    }
    task.delayedAt = delayedAt;
    /**
     * Retrun all delayed jobs, orginized by the timetsamp at where they are to run at.
     * Note: This is a very slow command.
     * Will throw an error if redis cannot be reached.
     */
    async function allDelayed() {
        return index_1.api.resque.queue.allDelayed();
    }
    task.allDelayed = allDelayed;
    /**
     * Retrun all workers registered by all members of this cluster.
     * Note: MultiWorker processors each register as a unique worker.
     * Will throw an error if redis cannot be reached.
     */
    async function workers() {
        return index_1.api.resque.queue.workers();
    }
    task.workers = workers;
    /**
     * What is a given worker working on?  If the worker is idle, 'started' will be returned.
     * Will throw an error if redis cannot be reached.
     */
    async function workingOn(workerName, queues) {
        return index_1.api.resque.queue.workingOn(workerName, queues);
    }
    task.workingOn = workingOn;
    /**
     * Return all workers and what job they might be working on.
     * Will throw an error if redis cannot be reached.
     */
    async function allWorkingOn() {
        return index_1.api.resque.queue.allWorkingOn();
    }
    task.allWorkingOn = allWorkingOn;
    /**
     * How many jobs are in the failed queue.
     * Will throw an error if redis cannot be reached.
     */
    async function failedCount() {
        return index_1.api.resque.queue.failedCount();
    }
    task.failedCount = failedCount;
    /**
     * Retrieve the details of failed jobs between start and stop (0-indexed).
     * Will throw an error if redis cannot be reached.
     */
    async function failed(start, stop) {
        return index_1.api.resque.queue.failed(start, stop);
    }
    task.failed = failed;
    /**
     * Remove a specific job from the failed queue.
     * Will throw an error if redis cannot be reached.
     */
    async function removeFailed(failedJob) {
        return index_1.api.resque.queue.removeFailed(failedJob);
    }
    task.removeFailed = removeFailed;
    /**
     * Remove a specific job from the failed queue, and retry it by placing it back into its original queue.
     * Will throw an error if redis cannot be reached.
     */
    async function retryAndRemoveFailed(failedJob) {
        return index_1.api.resque.queue.retryAndRemoveFailed(failedJob);
    }
    task.retryAndRemoveFailed = retryAndRemoveFailed;
    /**
     * If a worker process crashes, it will leave its state in redis as "working".
     * You can remove workers from redis you know to be over, by specificing an age which would make them too old to exist.
     * This method will remove the data created by a 'stuck' worker and move the payload to the error queue.
     * However, it will not actually remove any processes which may be running.  A job *may* be running that you have removed.
     * Will throw an error if redis cannot be reached.
     */
    async function cleanOldWorkers(age) {
        return index_1.api.resque.queue.cleanOldWorkers(age);
    }
    task.cleanOldWorkers = cleanOldWorkers;
    /**
     * Ensures that a task which has a frequency is either running, or already enqueued.
     * This is run automatically at boot for all tasks which have a frequency, via `api.tasks.enqueueAllRecurrentTasks`.
     * Will throw an error if redis cannot be reached.
     */
    async function enqueueRecurrentTask(taskName) {
        const thisTask = index_1.api.tasks.tasks[taskName];
        if (thisTask.frequency > 0) {
            await task.del(thisTask.queue, taskName);
            await task.delDelayed(thisTask.queue, taskName);
            await task.enqueueIn(thisTask.frequency, taskName, {});
            index_1.log(`re-enqueued recurrent job ${taskName}`, index_1.config.tasks.schedulerLogging.reEnqueue);
        }
    }
    task.enqueueRecurrentTask = enqueueRecurrentTask;
    /**
     * This is run automatically at boot for all tasks which have a frequency, calling `api.tasks.enqueueRecurrentTask`
     * Will throw an error if redis cannot be reached.
     */
    async function enqueueAllRecurrentTasks() {
        const jobs = [];
        const loadedTasks = [];
        Object.keys(index_1.api.tasks.tasks).forEach(taskName => {
            const thisTask = index_1.api.tasks.tasks[taskName];
            if (thisTask.frequency > 0) {
                jobs.push(async () => {
                    const toRun = await task.enqueue(taskName, {});
                    if (toRun === true) {
                        index_1.log(`enqueuing periodic task: ${taskName}`, index_1.config.tasks.schedulerLogging.enqueue);
                        loadedTasks.push(taskName);
                    }
                });
            }
        });
        await index_1.utils.asyncWaterfall(jobs);
        return loadedTasks;
    }
    task.enqueueAllRecurrentTasks = enqueueAllRecurrentTasks;
    /**
     * Stop a task with a frequency by removing it from all possible queues.
     * Will throw an error if redis cannot be reached.
     */
    async function stopRecurrentTask(taskName) {
        // find the jobs in either the normal queue or delayed queues
        const thisTask = index_1.api.tasks.tasks[taskName];
        if (thisTask.frequency > 0) {
            let removedCount = 0;
            const count = await task.del(thisTask.queue, thisTask.name, null, 1);
            removedCount = removedCount + count;
            const timestamps = await task.delDelayed(thisTask.queue, thisTask.name, null);
            removedCount = removedCount + timestamps.length;
            return removedCount;
        }
    }
    task.stopRecurrentTask = stopRecurrentTask;
    /**
     * Return wholistic details about the task system, including failures, queues, and workers.
     * Will throw an error if redis cannot be reached.
     */
    async function details() {
        const details = { queues: {}, workers: {}, stats: null };
        details.workers = await task.allWorkingOn();
        details.stats = await task.stats();
        const queues = await index_1.api.resque.queue.queues();
        for (const i in queues) {
            const queue = queues[i];
            const length = await index_1.api.resque.queue.length(queue);
            details.queues[queue] = { length: length };
        }
        return details;
    }
    task.details = details;
    async function addMiddleware(middleware) {
        if (!middleware.name) {
            throw new Error("middleware.name is required");
        }
        if (!middleware.priority) {
            middleware.priority = index_1.config.general.defaultMiddlewarePriority;
        }
        middleware.priority = Number(middleware.priority);
        index_1.api.tasks.middleware[middleware.name] = middleware;
        if (middleware.global === true) {
            index_1.api.tasks.globalMiddleware.push(middleware.name);
            index_1.utils.sortGlobalMiddleware(index_1.api.tasks.globalMiddleware, index_1.api.tasks.middleware);
        }
        index_1.api.tasks.loadTasks(true);
    }
    task.addMiddleware = addMiddleware;
})(task = exports.task || (exports.task = {}));
